import json
from typing import Any
from typing import override

from pyalab.plate import Plate

from .base import Step
from .base import ul_to_xml


class Transfer(Step):
    """Simple transfer from one column to another."""

    type = "Transfer"
    source: Plate
    """The source plate to aspirate from."""
    destination: Plate
    """The destination plate to dispense into."""
    source_section_index: int | None = None
    """The section index on the Deck of the source plate."""
    source_column_index: int
    """The column index to aspirate from."""
    destination_section_index: int | None = None
    """The section index on the Deck of the destination plate."""
    destination_column_index: int
    """The column index to dispense into."""
    volume: float
    """The volume to transfer."""

    @override
    def _add_value_groups(self) -> None:
        source_info: list[dict[str, Any]] = [
            {
                "Wells": [
                    {
                        "Item1": self.source_column_index,
                        "Item2": 0,  # TODO: handle row index
                    }
                ],
                "DeckSection": self.source_section_index,
                "SubSection": -1,  # TODO: figure out what subsection means
                "Spacing": 900,  # TODO: handle spacing other than 96-well plate
                "DeckId": "00000000-0000-0000-0000-000000000000",  # TODO: figure out if this has any meaning
                "WorkingDirectionExtended": 0,  # TODO: figure out what this is
                "WorkingDirectionOld": "false",  # TODO: figure out what this is
            }
        ]
        target_info: list[dict[str, Any]] = [
            {
                "Wells": [
                    {
                        "Item1": self.destination_column_index,
                        "Item2": 0,  # TODO: handle row index
                    }
                ],
                "DeckSection": self.destination_section_index,
                "SubSection": -1,  # TODO: figure out what subsection means
                "Spacing": 900,  # TODO: handle spacing other than 96-well plate
                "DeckId": "00000000-0000-0000-0000-000000000000",  # TODO: figure out if this has any meaning
                "WorkingDirectionExtended": 0,  # TODO: figure out what this is
                "WorkingDirectionOld": "false",  # TODO: figure out what this is
            }
        ]

        self._add_value_group(
            group_name="Source",
            values=[
                ("MultiSelection", json.dumps(source_info)),
                (
                    "WellOffsets",
                    json.dumps(
                        [
                            {
                                "DeckSection": self.source_section_index,
                                "SubSection": -1,
                                "OffsetX": 0,
                                "OffsetY": 0,
                            }
                        ]
                    ),
                ),
            ],
        )
        self._add_value_group(
            group_name="Target",
            values=[
                ("MultiSelection", json.dumps(target_info)),
                (
                    "WellOffsets",
                    json.dumps(
                        [
                            {
                                "DeckSection": self.destination_section_index,
                                "SubSection": -1,
                                "OffsetX": 0,
                                "OffsetY": 0,
                            }
                        ]
                    ),
                ),
            ],
        )
        self._add_value_group(
            group_name="Pipetting",
            values=[
                ("ExtraVolumePercentage", str(0)),
                ("NumberOfReactions", str(1)),  # TODO: handle multiple transfers in a single Step
                (
                    "DispenseVolume",
                    json.dumps(
                        [
                            {
                                "Well": {"Item1": self.destination_column_index, "Item2": 0},
                                "DeckSection": self.destination_section_index,
                                "SubSection": -1,
                                "Volume": ul_to_xml(self.volume),
                                "TipID": self.tip_id,
                                "Multiplier": 1,
                                "TotalVolume": ul_to_xml(
                                    self.volume
                                ),  # TODO: figure out when/if this needs to differ from Volume
                            }
                        ]
                    ),
                ),
                (
                    "TipTypePipettingConfiguration",
                    json.dumps(
                        [
                            {
                                "FirstDispenseVolume": 0,
                                "LastDispenseVolume": 0,
                                "Airgap": False,
                                "AirgapVolume": 0,
                                "AspirationSpeed": 8,
                                "DispenseSpeed": 8,
                                "TipID": self.tip_id,
                            }
                        ]
                    ),
                ),
                ("AspirationDelay", str(0)),
                ("DispenseDelay", str(0)),
                ("KeepPostDispense", json.dumps(obj=False)),
                ("LastDispenseType", json.dumps(obj=True)),
                ("LastAspirationBackTo", '"Common_No"'),
                ("VolumeConfigType", json.dumps(obj=True)),
                ("DispenseType", json.dumps(obj=False)),
                ("SlowLiquidExitAsp", json.dumps(obj=False)),
                ("SlowLiquidExitDisp", json.dumps(obj=False)),
            ],
        )
        self._add_value_group(
            group_name="Aspiration",
            values=[
                (
                    "Heights",
                    json.dumps(
                        [
                            {
                                "Well": {"Item1": self.source_column_index, "Item2": 0},
                                "DeckSection": self.source_section_index,
                                "SubSection": -1,
                                "StartHeight": 325,  # TODO: figure out how these height values are determined
                                "EndHeight": 325,
                                "TipID": self.tip_id,
                            }
                        ]
                    ),
                ),
                ("TipTravel", json.dumps(obj=False)),
                (
                    "SectionHeightConfig",
                    json.dumps(
                        [
                            {
                                "DeckSection": self.source_section_index,
                                "SubSection": -1,
                                "HeightConfigType": True,
                                "WellBottomOffset": 0,
                            }
                        ]
                    ),
                ),
                (
                    "TipTypeHeightConfiguration",
                    json.dumps(
                        [
                            {
                                "DeckSection": self.source_section_index,
                                "SubSection": -1,
                                "WellBottomOffset": 200,
                                "TipID": self.tip_id,
                            }
                        ]
                    ),
                ),
            ],
        )
        self._add_value_group(
            group_name="Dispense",
            values=[
                (
                    "Heights",
                    json.dumps(
                        [
                            {
                                "Well": {"Item1": self.destination_column_index, "Item2": 0},
                                "DeckSection": self.destination_section_index,
                                "SubSection": -1,
                                "StartHeight": 325,  # TODO: figure out how these height values are determined
                                "EndHeight": 325,
                                "TipID": self.tip_id,
                            }
                        ]
                    ),
                ),
                ("TipTravel", json.dumps(obj=False)),
                (
                    "SectionHeightConfig",
                    json.dumps(
                        [
                            {
                                "DeckSection": self.destination_section_index,
                                "SubSection": -1,
                                "HeightConfigType": True,
                                "WellBottomOffset": 0,
                            }
                        ]
                    ),
                ),
                (
                    "TipTypeHeightConfiguration",
                    json.dumps(
                        [
                            {
                                "DeckSection": self.destination_section_index,
                                "SubSection": -1,
                                "WellBottomOffset": 200,
                                "TipID": self.tip_id,
                            }
                        ]
                    ),
                ),
            ],
        )
        self._add_value_group(
            group_name="Tips",
            values=[
                ("PreWetting", json.dumps(obj=False)),
                ("PreWettingCycles", json.dumps(obj=3)),
                ("TipChange", '"TipChange_ModeA"'),
                ("TipEjectionType", json.dumps(obj=True)),
            ],
        )
        self._add_value_group(
            group_name="SourceMix",
            values=[
                ("MixActive", json.dumps(obj=False)),
                (
                    "TipTypeMixConfiguration",
                    json.dumps(
                        obj=[
                            {
                                "MixSpeed": 8,
                                "TipID": self.tip_id,
                            }
                        ]
                    ),
                ),
                ("MixPause", json.dumps(obj=0)),
                (
                    "SectionMixVolume",
                    json.dumps(
                        obj=[
                            {
                                "Well": {"Item1": self.source_column_index, "Item2": 0},
                                "DeckSection": self.source_section_index,
                                "SubSection": -1,
                                "Volume": 5000,  # TODO: implement mixing volume
                                "TipID": self.tip_id,
                                "Multiplier": 1,
                                "TotalVolume": 5000,  # TODO: figure out when/if this needs to differ from Volume
                            }
                        ]
                    ),
                ),
                ("MixCycles", json.dumps(obj=3)),
                ("BlowOut", json.dumps(obj=False)),
                ("TipTravel", json.dumps(obj=False)),
                (
                    "SectionHeightConfig",
                    json.dumps(
                        obj=[
                            {
                                "DeckSection": self.source_section_index,
                                "SubSection": -1,
                                "HeightConfigType": True,
                                "WellBottomOffset": 0,
                            }
                        ]
                    ),
                ),
                ("VolumeConfigType", json.dumps(obj=True)),
                (
                    "Heights",
                    json.dumps(
                        obj=[
                            {
                                "Well": {"Item1": self.source_column_index, "Item2": 0},
                                "DeckSection": self.source_section_index,
                                "SubSection": -1,
                                "StartHeight": 325,  # TODO: figure out how these height values are determined
                                "EndHeight": 0,
                                "TipID": self.tip_id,
                            }
                        ]
                    ),
                ),
                ("MixBeforeEachAspiration", json.dumps(obj=False)),
            ],
        )
        self._add_value_group(
            group_name="TargetMix",
            values=[
                ("MixActive", json.dumps(obj=False)),
                (
                    "TipTypeMixConfiguration",
                    json.dumps(
                        obj=[
                            {
                                "MixSpeed": 8,
                                "TipID": self.tip_id,
                            }
                        ]
                    ),
                ),
                ("MixPause", json.dumps(obj=0)),
                (
                    "SectionMixVolume",
                    json.dumps(
                        obj=[
                            {
                                "Well": {"Item1": self.destination_column_index, "Item2": 0},
                                "DeckSection": self.destination_section_index,
                                "SubSection": -1,
                                "Volume": 5000,  # TODO: implement mixing volume
                                "TipID": self.tip_id,
                                "Multiplier": 1,
                                "TotalVolume": 5000,  # TODO: figure out when/if this needs to differ from Volume
                            }
                        ]
                    ),
                ),
                ("MixCycles", json.dumps(obj=3)),
                ("BlowOut", json.dumps(obj=False)),
                ("TipTravel", json.dumps(obj=False)),
                (
                    "SectionHeightConfig",
                    json.dumps(
                        obj=[
                            {
                                "DeckSection": self.destination_section_index,
                                "SubSection": -1,
                                "HeightConfigType": True,
                                "WellBottomOffset": 0,
                            }
                        ]
                    ),
                ),
                ("VolumeConfigType", json.dumps(obj=True)),
                (
                    "Heights",
                    json.dumps(
                        obj=[
                            {
                                "Well": {"Item1": self.destination_column_index, "Item2": 0},
                                "DeckSection": self.destination_section_index,
                                "SubSection": -1,
                                "StartHeight": 325,  # TODO: figure out how these height values are determined
                                "EndHeight": 0,
                                "TipID": self.tip_id,
                            }
                        ]
                    ),
                ),
                ("MixBeforeEachAspiration", json.dumps(obj=False)),
                ("SkipFirst", json.dumps(obj=False)),
            ],
        )
        self._add_value_group(
            group_name="TipTouchTarget",
            values=[
                ("TipTouchActive", json.dumps(obj=False)),
                (
                    "SectionTipTouch",
                    json.dumps(
                        obj=[
                            {
                                "DeckSection": self.destination_section_index,
                                "SubSection": -1,
                                "Type": False,
                                "Height": 1406,  # TODO: implement tip touch
                                "Distance": 225,
                            }
                        ]
                    ),
                ),
            ],
        )
        self._add_value_group(
            group_name="Various",
            values=[
                ("SpeedX", str(10)),
                ("SpeedY", str(10)),
                ("SpeedZ", str(10)),
                ("IsStepActive", json.dumps(obj=True)),
            ],
        )



import json
import re
import uuid
from pathlib import Path
from xml.dom import minidom

from lxml import etree
from pydantic import BaseModel
from pydantic import Field

from .deck import DeckLayout
from .pipette import Pipette
from .pipette import Tip
from .plate import Plate
from .steps import Step


class LabwareNotInDeckLayoutError(Exception):
    def __init__(self, plate: Plate):
        super().__init__(f"Could not find {plate.name} (called {plate.display_name}) in the deck layout")


class Program(BaseModel):
    deck_layouts: list[DeckLayout] = Field(min_length=1)  # TODO: validate that all layouts use the same base Deck
    display_name: str  # TODO: validate length and character classes
    description: str  # TODO: validate length and character classes
    pipette: Pipette
    tip: Tip
    steps: list[Step] = Field(default_factory=list)

    def add_step(self, step: Step) -> None:
        step.set_tip(self.tip)
        self.steps.append(step)

    def get_section_index_for_plate(self, plate: Plate) -> int:
        # TODO: support multiple deck layouts
        first_deck_layout = self.deck_layouts[0]
        for deck_position, iter_plate in first_deck_layout.labware.items():
            if iter_plate == plate:
                return deck_position.section_index(first_deck_layout.deck)

        raise LabwareNotInDeckLayoutError(plate)

    def generate_xml(self) -> str:
        config_version = 4
        data_version = 9
        root = etree.Element(
            "AssistConfig",
            nsmap={"xsd": "http://www.w3.org/2001/XMLSchema", "xsi": "http://www.w3.org/2001/XMLSchema-instance"},
            UniqueIdentifier=str(uuid.uuid4()),
            Version=str(config_version),
        )

        for element_name, text_value in [
            ("MigrationIdentifier", str(uuid.uuid4())),
            ("CreatedWith", f"PyaLab for VIALAB v3.4.0.0, config v{config_version}, data v{data_version}"),
            ("CreatedBy", "UnknownUser"),
            ("MigrationHistory", ""),
            ("DataVersion", str(data_version)),
            ("DisplayNameOnPipette", self.display_name),
            ("Description", self.description),
        ]:
            etree.SubElement(root, element_name).text = text_value

        root.append(self.pipette.create_xml_for_program())
        root.append(self.tip.create_xml_for_program())
        tips_node = etree.SubElement(root, "Tips")
        # TODO: figure out how to handle multiple tip types, likely for the D-ONE
        tips_node.append(self.tip.create_xml_for_program())

        # TODO: handle multiple deck layouts
        first_deck_layout = self.deck_layouts[0]
        root.append(first_deck_layout.create_xml_for_program(layout_num=1))
        decks_node = etree.SubElement(root, "AllDecks")
        decks_node.append(first_deck_layout.create_xml_for_program(layout_num=1))

        steps_node = etree.SubElement(root, "Steps")
        for step in self.steps:
            steps_node.append(step.create_xml_for_program())

        global_parameters_node = etree.SubElement(root, "GlobalParameters", attrib={"Key": "Global"})
        global_parameters_value_node = etree.SubElement(global_parameters_node, "Values")
        for key, value in [
            ("ClearanceHeight", 800),
            ("SectionOffsets", "null"),
            ("DisplayTipEjectionOptions", "true"),
            ("AfterTipEjectMonitoring", "true"),
            ("AfterTipLoadMonitoring", "false"),
            ("BeforeTipEjectMonitoring", "true"),
            (
                "TipTypeRequiredTips",
                json.dumps(
                    {
                        str(
                            self.tip.tip_id
                        ): 0  # there seems to be no negative impact of not calculating the required tips, ViaLab will do it automatically when the program is first loaded
                    }
                ),
            ),
            ("WasteAsTargetOption", "false"),
            ("LabwareReintegration", "false"),
            ("CopyHeightAdjustment", "false"),
            ("WellBottomMinHeight", 200),
            ("CollisionAvoidanceOffset", 0),
            ("CollisionDetection", "true"),
        ]:
            etree.SubElement(global_parameters_value_node, "Value", attrib={"Key": key}).text = (
                etree.CDATA(str(value)) if '"' in str(value) else str(value)
            )

        _ = etree.SubElement(root, "ChangedDate").text = (
            "2024-12-17T16:27:27.0715524-05:00"  # TODO: make this real time
        )
        _ = etree.SubElement(root, "LastChangeUser").text = "UnknownUser"

        xml_string = etree.tostring(root, xml_declaration=True, encoding="utf-8")
        xml_str = minidom.parseString(xml_string).toprettyxml(indent="  ")  # noqa: S318 # it is safe to parse this string because it is generated by the program
        xml_str_cleaned = re.sub(r"\n\s*\n", "\n", xml_str)
        return xml_str_cleaned.replace(
            '<?xml version="1.0" ?>', '<?xml version="1.0" encoding="utf-8"?>'
        )  # TODO: figure out why the encoding argument to `tostring` isn't working as expected

    def dump_xml(self, file_path: Path) -> None:
        xml_str = self.generate_xml()
        _ = file_path.write_text(xml_str)



from pathlib import Path

PATH_TO_PACKAGE_ROOT = Path(__file__).absolute().resolve().parent
assert str(PATH_TO_PACKAGE_ROOT).endswith(
    "pyalab"
), f"Sanity check failed, path was not to package root, it was: {PATH_TO_PACKAGE_ROOT}"
PATH_TO_INCLUDED_XML_FILES = PATH_TO_PACKAGE_ROOT / "vendor_files" / "integra_library"



import json
import uuid
from abc import ABC
from abc import abstractmethod
from typing import ClassVar

from lxml import etree
from lxml.etree import _Element
from pydantic import BaseModel

from pyalab.pipette import Tip


def ul_to_xml(volume: float) -> int:
    # ViaLab uses 0.01 uL as the base unit for volume, so convert from uL
    return int(round(volume * 100, 0))


SPECIAL_CHARS = ('"', "[", "]", "{", "}")


class Step(BaseModel, ABC):
    type: ClassVar[str]
    _tip: Tip | None = None

    def set_tip(self, tip: Tip) -> None:
        self._tip = tip

    @property
    def tip(self) -> Tip:
        assert self._tip is not None
        return self._tip

    @property
    def tip_id(self) -> int:
        return self.tip.tip_id

    def create_xml_for_program(self) -> _Element:
        root = etree.Element("Step")
        for name, value in [
            ("Type", self.type),
            ("IsEnabled", "true"),
            ("ID", str(uuid.uuid4())),
            ("IsNew", json.dumps(obj=False)),
            (
                "DeckID",
                "00000000-0000-0000-0000-000000000000",
            ),  # TODO: figure out what this is and if it needs to be changed
        ]:
            etree.SubElement(root, name).text = value

        self._value_groups_node = etree.SubElement(root, "ValueGroups")
        self._add_value_groups()
        return root

    @abstractmethod
    def _add_value_groups(self) -> None: ...

    def _add_value_group(self, *, group_name: str, values: list[tuple[str, str]]) -> None:
        group_node = etree.SubElement(self._value_groups_node, "ValueGroup", attrib={"Key": group_name})
        values_node = etree.SubElement(group_node, "Values")
        for name, value in values:
            is_c_data_needed = any(char in value for char in SPECIAL_CHARS)
            etree.SubElement(values_node, "Value", attrib={"Key": name}).text = (
                etree.CDATA(value) if is_c_data_needed else value
            )



import json
from typing import Any
from typing import override

from pydantic import Field

from pyalab.plate import Plate

from .base import Step
from .base import ul_to_xml


class SetVolume(Step):
    type = "ManualFilling"
    plate: Plate
    section_index: int | None = None
    column_index: int
    volume: float = Field(ge=0)

    @override
    def _add_value_groups(self) -> None:
        volume_info: list[dict[str, Any]] = [
            {
                "WellCoordinates": [
                    {
                        "Item1": self.column_index,
                        "Item2": 0,  # TODO: handle row index
                    }
                ],
                "Volume": ul_to_xml(self.volume),
                "Section": self.section_index,
                "SubSection": -1,  # TODO: figure out what subsection means
                "Spacing": 900,  # TODO: handle spacing other than 96-well plate
                "ColorIndex": 1,  # TODO: figure out if/when this changes
                "DeckId": "00000000-0000-0000-0000-000000000000",  # TODO: figure out if this has any meaning
            }
        ]

        self._add_value_group(
            group_name="ManualVolume",
            values=[
                ("ManualVolume", json.dumps(volume_info)),
                ("MessageType", "null"),
                ("Message1", "null"),
                ("Message2", "null"),
                ("Message3", "null"),
                ("ShowMessageOnPipette", "false"),
            ],
        )


class SetInitialVolume(SetVolume):
    type = "ManualFilling_First"



import re
from enum import Enum
from pathlib import Path
from typing import ClassVar

from lxml import etree
from lxml.etree import _Element
from pydantic import BaseModel

from .constants import PATH_TO_INCLUDED_XML_FILES


class LibraryComponentType(Enum):
    DECK = "Deck"
    PLATE = "Plate"
    RESERVOIR = "Reservoir"
    PIPETTE = "Pipette"
    TIP = "Tip"


class IntegraLibraryObjectNotFoundError(OSError):
    def __init__(self, *, component_type: LibraryComponentType, name: str, paths_searched: list[Path]):
        self.type = component_type
        self.name = name
        super().__init__(f"Could not find {component_type.value} with name {name} while looking in {paths_searched}")


class LibraryComponent(BaseModel, frozen=True):
    type: ClassVar[LibraryComponentType]
    name: str
    xml_file_version: str | None = None

    def load_xml(self) -> _Element:
        directory = PATH_TO_INCLUDED_XML_FILES / self.type.value
        xml_files = directory.glob("*.xml")
        regex_pattern = re.compile(rf"{self.name}\ V\d+\.xml")
        matched_files = [file for file in xml_files if regex_pattern.match(file.name)]
        if len(matched_files) == 0:
            raise IntegraLibraryObjectNotFoundError(
                component_type=self.type, name=self.name, paths_searched=[directory]
            )

        assert len(matched_files) == 1  # TODO: handle multiple versions in the library...
        file = matched_files[0]
        parser = etree.XMLParser(no_network=True, recover=False)
        tree = etree.parse(file, parser)  # noqa: S320 # using this custom parser should mitigate security concerns about untrusted XML files. And defusedxml is end of life anyway
        root = tree.getroot()
        assert isinstance(root, _Element), f"Expected root to be an Element, but got type {type(root)} for {root}"
        return root

    def create_xml_for_program(self) -> _Element:
        is_content = self.type in [LibraryComponentType.PLATE, LibraryComponentType.RESERVOIR]
        root = etree.Element(
            "Content"
            if is_content
            else self.type.value,  # TODO: confirm that all object types use the file directory as the XML tag name too
            Version=str(1),  # TODO: confirm that no objects are using version other than 1
        )
        if is_content:
            root.set(
                etree.QName("http://www.w3.org/2001/XMLSchema-instance", "type"),
                self.type.value,  # TODO: confirm that all object types use the file directory as the xsi:type too
            )

        for subelement in self.load_xml():
            root.append(subelement)
        return root



from enum import Enum
from functools import cached_property
from typing import ClassVar
from typing import Literal

from lxml import etree
from lxml.etree import _Element
from pydantic import BaseModel

from .integra_xml import LibraryComponent
from .integra_xml import LibraryComponentType
from .plate import Plate


class StandardDeckNames(Enum):
    THREE_POSITION = "3 Position Universal Deck"
    FOUR_POSITION = "4 Position Portrait Deck"


class Deck(LibraryComponent, frozen=True):
    type: ClassVar[LibraryComponentType] = LibraryComponentType.DECK


class DeckPositionNotFoundError(Exception):
    def __init__(self, *, deck_name: str, deck_position_name: str, width: int, length: int):
        super().__init__(
            f"Could not find a position in the Deck {deck_name} with name {deck_position_name} of width {width} and length {length}"
        )


class DeckPosition(BaseModel, frozen=True):
    name: Literal["A", "B", "C", "D"]
    width: float
    length: float

    # the XML encodes the dimension in units of 0.01 mm, but our standard units are in mm
    @cached_property
    def xml_width(self) -> int:
        return int(round(self.width * 100, 0))

    @cached_property
    def xml_length(self) -> int:
        return int(round(self.length * 100, 0))

    def section_index(self, deck: Deck) -> int:
        root = deck.load_xml()
        for idx, section in enumerate(root.findall("./Sections/Section")):
            name = section.find("Name")
            width = section.find("Width")
            length = section.find("Length")
            assert width is not None
            assert length is not None
            assert name is not None

            if name.text == self.name and width.text == str(self.xml_width) and length.text == str(self.xml_length):
                return idx  # TODO: confirm that Integra does not allow any duplicates inherently

        raise DeckPositionNotFoundError(
            deck_name=deck.name, deck_position_name=self.name, width=self.xml_width, length=self.xml_length
        )
        # TODO: figure out if CreationOrderIndex is important to be changed or not


class DeckPositions(Enum):
    B_PLATE_LANDSCAPE = DeckPosition(name="B", length=128.2, width=86)
    C_PLATE_LANDSCAPE = DeckPosition(name="C", length=128.2, width=86)


class DeckLayout(BaseModel):
    deck: Deck
    labware: dict[DeckPosition, Plate]
    name: str = ""

    def create_xml_for_program(self, *, layout_num: int) -> _Element:
        name = self.name if self.name != "" else f"Labware Layout {layout_num}"

        root = self.deck.create_xml_for_program()
        _ = etree.SubElement(root, "NameInProcess").text = name
        for deck_position, plate in self.labware.items():
            section_idx = deck_position.section_index(self.deck)

            for idx, section in enumerate(root.findall("./Sections/Section")):
                if idx == section_idx:
                    plate_xml = plate.create_xml_for_program()
                    children = list(section)
                    for index, child in enumerate(list(section)):
                        if child.tag == "IsWaste":
                            # Insert the new element right after <IsWaste>
                            children.insert(index + 1, plate_xml)
                            break
                    else:
                        raise NotImplementedError(
                            "Could not find <IsWaste> element in the section...this should never happen so there's no implementation to handle it"
                        )
                    # Clear existing children and re-insert the updated list
                    section.clear()
                    section.extend(children)

                    break
            else:
                raise NotImplementedError(
                    f"Could not find section with index {section_idx} in the deck XML...this should never happen so there's no implementation to handle it"
                )

        return root



import uuid
from typing import override

from lxml import etree
from lxml.etree import _Element
from pydantic import Field

from .integra_xml import LibraryComponent
from .integra_xml import LibraryComponentType


class Plate(LibraryComponent, frozen=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    type = LibraryComponentType.PLATE
    display_name: str = ""  # TODO: If left as blank, then set the display name to the name of the plate type # TODO: validate length and character class requirements

    @override
    def create_xml_for_program(self) -> _Element:
        root = super().create_xml_for_program()
        etree.SubElement(
            root, "NameInProcess"
        ).text = f"{self.display_name}!1"  # TODO: figure out why they all end in `!1`
        return root



from functools import cached_property

from .integra_xml import LibraryComponent
from .integra_xml import LibraryComponentType


class Pipette(LibraryComponent, frozen=True):
    type = LibraryComponentType.PIPETTE
    name: str


class Tip(LibraryComponent, frozen=True):
    type = LibraryComponentType.TIP
    name: str

    @cached_property
    def tip_id(self) -> int:
        root = self.load_xml()
        tip_id_node = root.find(".//TipID")
        assert tip_id_node is not None
        tip_id_text = tip_id_node.text
        assert tip_id_text is not None
        return int(tip_id_text)



from .constants import PATH_TO_INCLUDED_XML_FILES
from .deck import Deck
from .deck import DeckLayout
from .deck import DeckPosition
from .deck import DeckPositionNotFoundError
from .deck import DeckPositions
from .deck import StandardDeckNames
from .integra_xml import IntegraLibraryObjectNotFoundError
from .integra_xml import LibraryComponent
from .integra_xml import LibraryComponentType
from .pipette import Pipette
from .pipette import Tip
from .plate import Plate
from .program import LabwareNotInDeckLayoutError
from .program import Program
from .steps import SetInitialVolume
from .steps import SetVolume
from .steps import Step
from .steps import Transfer

__all__ = [
    "PATH_TO_INCLUDED_XML_FILES",
    "Deck",
    "DeckLayout",
    "DeckPosition",
    "DeckPositionNotFoundError",
    "DeckPositions",
    "IntegraLibraryObjectNotFoundError",
    "LabwareNotInDeckLayoutError",
    "LibraryComponent",
    "LibraryComponentType",
    "Pipette",
    "Plate",
    "Program",
    "SetInitialVolume",
    "SetVolume",
    "StandardDeckNames",
    "Step",
    "Tip",
    "Transfer",
]
